/*
          Copyright (C) 1995 - 1996, RSNA and Washington University

          The software and supporting documentation for the Radiological
          Society of North America (RSNA) 1993 - 1996 Digital Imaging and
          Communications in Medicine (DICOM) Demonstration were developed
          at the
                  Electronic Radiology Laboratory
                  Mallinckrodt Institute of Radiology
                  Washington University School of Medicine
                  510 S. Kingshighway Blvd.
                  St. Louis, MO 63110
          as part of the 1993 - 1996 DICOM Central Test Node project for, and
          under contract with, the Radiological Society of North America.

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER RSNA NOR
          WASHINGTON UNIVERSITY MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          jointly owned by RSNA and Washington University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include the three paragraphs of
          the copyright notice.
*/
/*
+-+-+-+-+-+-+-+-+-
*/
/*
**		     Electronic Radiology Laboratory
**		   Mallinckrodt Institute of Radiology
**		Washington University School of Medicine
**
** Module Name(s):
** Author, Date:	Steve Moore, 30-Jun-96
** Intent:		Provide common abstractions needed for operations
**			in a multi-threaded environment.
** Last Update:		$Author: smm $, $Date: 1998/06/10 21:54:37 $
** Source File:		$RCSfile: ctnthread.c,v $
** Revision:		$Revision: 1.4 $
** Status:		$State: Exp $
*/

static char rcsid[] = "$Revision: 1.4 $ $RCSfile: ctnthread.c,v $";

#include <stdio.h>
#ifdef GCCSUNOS
#include <sys/types.h>
#endif
#include "dicom.h"
#include "condition.h"
#include "ctnthread.h"

char *THR_Message(CONDITION cond);

#ifdef CTN_USE_THREADS
#ifdef _MSC_VER
#include <windows.h>
LPCTSTR mutexName = "CTN Mutex";
HANDLE hMutex = NULL;
#else
#include <synch.h>
static mutex_t mutex;
#endif
#endif

#ifdef CTN_USE_THREADS
static CTNBOOLEAN initialized = FALSE;
#endif

CONDITION
THR_Init()
{
#ifdef CTN_USE_THREADS
#ifdef _MSC_VER
    if (initialized)
	return THR_NORMAL;

    hMutex = CreateMutex(NULL, FALSE, mutexName);
    initialized = TRUE;
    return THR_NORMAL;

#else

    int cond;

    if (initialized)
	return THR_NORMAL;

    cond = mutex_init(&mutex, USYNC_THREAD, NULL);
    if (cond != 0) {
	return COND_PushCondition(THR_GENERICFAILURE,
				  THR_Message(THR_GENERICFAILURE),
				"Failed on call to mutex_init", "THR_Init");
    }
    initialized = TRUE;
    return THR_NORMAL;
#endif
#else
    return THR_NORMAL;
#endif
}

CONDITION
THR_Shutdown()
{
#ifdef CTN_USE_THREADS
#ifdef _MSC_VER
    if (!initialized) {
	return COND_PushCondition(THR_NOTINITIALIZED,
			   THR_Message(THR_NOTINITIALIZED), "THR_Shutdown");
    }
    CloseHandle(hMutex);
    return THR_NORMAL;

#else
    int cond;

    if (!initialized) {
	return COND_PushCondition(THR_NOTINITIALIZED,
			   THR_Message(THR_NOTINITIALIZED), "THR_Shutdown");
    }
    cond = mutex_destroy(&mutex);
    if (cond != 0) {
	return COND_PushCondition(THR_GENERICFAILURE,
				  THR_Message(THR_GENERICFAILURE),
			 "Failed on call to mutex_destroy", "THR_Shutdown");
    }
    return THR_NORMAL;
#endif
#else
    return THR_NORMAL;
#endif
}

CONDITION
THR_ObtainMutex()
{
#ifdef CTN_USE_THREADS
#ifdef _MSC_VER
    HANDLE hTmpMutex = OpenMutex(MUTEX_ALL_ACCESS, FALSE, mutexName);
    WaitForSingleObject(hMutex, INFINITE);

    return THR_NORMAL;
#else
    int cond;

    if (!initialized) {
	return COND_PushCondition(THR_NOTINITIALIZED,
			THR_Message(THR_NOTINITIALIZED), "THR_ObtainMutex");
    }
    cond = mutex_lock(&mutex);
    if (cond != 0) {
	return COND_PushCondition(THR_GENERICFAILURE,
				  THR_Message(THR_GENERICFAILURE),
			 "Failed on call to mutex_lock", "THR_ObtainMutex");
    }
    return THR_NORMAL;
#endif
#else
    return THR_NORMAL;
#endif
}

CONDITION
THR_ReleaseMutex()
{
#ifdef CTN_USE_THREADS
#ifdef _MSC_VER
    ReleaseMutex(hMutex);
    return THR_NORMAL;

#else
    int cond;

    if (!initialized) {
	return COND_PushCondition(THR_NOTINITIALIZED,
			THR_Message(THR_NOTINITIALIZED), "THR_ObtainMutex");
    }
    cond = mutex_unlock(&mutex);
    if (cond != 0) {
	return COND_PushCondition(THR_GENERICFAILURE,
				  THR_Message(THR_GENERICFAILURE),
		      "Failed on call to mutex_unlock", "THR_ReleaseMutex");
    }
    return THR_NORMAL;
#endif
#else
    return THR_NORMAL;
#endif
}
